# miscellaneous useful tools
from __future__ import print_function, unicode_literals, absolute_import, division
import json
from six.moves import urllib
import six
import threading
import sys
import traceback
import os
import re

from astropy.io import fits
from astropy.io import ascii
from astropy.table import Table

from . import DriverError

from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet import reactor
from twisted.internet.task import deferLater


if not six.PY3:
    import tkFileDialog as filedialog
    from StringIO import StringIO
else:
    from tkinter import filedialog
    from io import StringIO


def async_sleep(secs):
    return deferLater(reactor, secs, lambda: None)


class ReadNGCTelemetry(object):
    """
    Class to field the telemetry sent by the NGC server

    Set the following attributes:
     ok      : True if telemetry is read OK
     err     : reason if telemetry not parsed OK
     state   : state of the camera. Possibilties are:
               'IDLE', 'BUSY', 'ERROR', 'ABORT', 'UNKNOWN'
     clocks  : whether the clock voltages are enabled. Possible values:
               'enabled', 'disabled'
     run     : current or last run number
    """

    def __init__(self, telemetry):
        """
        Parameters
        ----------
        telemetry : dict
            telemetry package from NGC
        """
        # Determine state of the camera
        self.ok = True
        self.err = ""
        if "system.subStateName" not in telemetry:
            self.err += "could not identify state\n"
            self.state = None
        else:
            self.state = telemetry["system.subStateName"]

        # determine state of clocks
        if "cldc_0.statusName" not in telemetry:
            self.ok = False
            self.err += "could not identify clock status\n"
            self.clocks = None
        else:
            self.clocks = telemetry["cldc_0.statusName"]

        # Find current run number (set it to 0 if we fail)
        newDataFileName = telemetry["exposure.newDataFileName"]
        exposure_state = telemetry["exposure.expStatusName"]
        pattern = "\D*(\d*).*.fits"
        try:
            run_number = int(re.match(pattern, newDataFileName).group(1))
            if exposure_state in ["success", "aborted", "inactive", "failure"]:
                self.run = run_number
            elif exposure_state == "integrating":
                self.run = run_number + 1
            else:
                raise ValueError("unknown exposure state {}".format(exposure_state))
        except (ValueError, IndexError, AttributeError):
            self.run = 0
            self.ok = False
            self.err += "cannot read run number"


def overlap(xl1, yl1, nx1, ny1, xl2, yl2, nx2, ny2):
    """
    Determines whether two windows overlap
    """
    return xl2 < xl1 + nx1 and xl2 + nx2 > xl1 and yl2 < yl1 + ny1 and yl2 + ny2 > yl1


@inlineCallbacks
def startNodding(g, data):
    nodPattern = data.get("appdata", {}).get("nodpattern", {})
    if g.cpars["telins_name"] == "GTC" and nodPattern:
        session = g.session
        if session is None:
            g.clog.warn("no WAMP session")
            returnValue(False)
        try:
            yield session.call("hipercam.gtc.rpc.gtc.start_nodding")
        except Exception as err:
            g.clog.warn("Failed to stop dither server")
            msg = err.error_message() if hasattr(err, "error_message") else str(err)
            g.clog.warn(msg)
            returnValue(False)
    returnValue(True)


@inlineCallbacks
def stopNodding(g):
    if g.cpars["telins_name"] == "GTC":
        session = g.session
        if session is None:
            g.clog.warn("no WAMP session")
            returnValue(False)
        try:
            yield session.call("hipercam.gtc.rpc.gtc.stop_nodding")
        except Exception as err:
            g.clog.warn("Failed to stop dither server")
            msg = err.error_message() if hasattr(err, "error_message") else str(err)
            g.clog.warn(msg)
            returnValue(False)
    returnValue(True)


def saveJSON(g, data, backup=False):
    """
    Saves the current setup to disk.

    g : hcam_drivers.globals.Container
    Container with globals

    data : dict
    The current setup in JSON compatible dictionary format.

    backup : bool
    If we are saving a backup on close, don't prompt for filename
    """
    if not backup:
        fname = filedialog.asksaveasfilename(
            defaultextension=".json",
            filetypes=[
                ("json files", ".json"),
            ],
            initialdir=g.cpars["app_directory"],
        )
    else:
        fname = os.path.join(os.path.expanduser("~/.hdriver"), "app.json")

    if not fname:
        g.clog.warn("Aborted save to disk")
        return False

    with open(fname, "w") as of:
        of.write(json.dumps(data, sort_keys=True, indent=4, separators=(",", ": ")))
    g.clog.info("Saved setup to" + fname)
    return True


@inlineCallbacks
def postJSON(g, data):
    """
    Posts the current setup to the camera and data servers.

    g : hcam_drivers.globals.Container
    Container with globals

    data : dict
    The current setup in JSON compatible dictionary format.
    """
    g.clog.debug("Entering postJSON")
    session = g.session
    if session is None:
        g.clog.warn("no WAMP session")
        returnValue(False)

    ok = True
    try:
        ok, status_msg = yield session.call("hipercam.ngc.rpc.load_setup", data)
    except Exception as err:
        status_msg = err.error_message() if hasattr(err, "error_message") else str(err)
        ok = False

    if not ok:
        g.clog.warn("Server response was not OK")
        g.rlog.warn("Error: " + status_msg)
        returnValue(False)

    # now try to setup nodding server if appropriate
    nodpattern = data.get("appdata", {}).get("nodpattern", {})
    if g.cpars["telins_name"] == "GTC" and nodpattern:
        try:
            ra_offsets = list(nodpattern["ra"])
            dec_offsets = list(nodpattern["dec"])
            yield session.call(
                "hipercam.gtc.rpc.load_nod_pattern", ra_offsets, dec_offsets
            )
            ok = True
        except Exception as err:
            status_msg = (
                err.error_message() if hasattr(err, "error_message") else str(err)
            )
            ok = False

        if not ok:
            g.clog.warn("Offset Server response was not OK")
            g.rlog.warn("Error: " + status_msg)
            returnValue(False)

    g.clog.debug("Leaving postJSON")
    returnValue(True)


@inlineCallbacks
def createJSON(g, full=True):
    """
    Create JSON compatible dictionary from current settings

    Parameters
    ----------
    g :  hcam_drivers.globals.Container
    Container with globals
    """
    data = dict()
    if "gps_attached" not in g.cpars:
        data["gps_attached"] = 1
    else:
        data["gps_attached"] = 1 if g.cpars["gps_attached"] else 0

    data["appdata"] = g.ipars.dumpJSON()
    data["user"] = g.rpars.dumpJSON()
    if full:
        data["hardware"] = g.ccd_hw.dumpJSON()
        data["tcs"] = g.info.dumpJSON()
        data["compo"] = g.compo_hw.dumpJSON()

        if g.cpars["telins_name"].lower() == "gtc":
            session = g.session
            if session is None:
                g.clog.warn("no WAMP session, not fetching telescope parameters")
            else:
                try:
                    telpars = yield session.call("hipercam.gtc.rpc.get_telescope_pars")
                    data["gtc_headers"] = telpars
                except Exception as err:
                    g.clog.warn("cannot get GTC headers from telescope server")
                    msg = (
                        err.error_message()
                        if hasattr(err, "error_message")
                        else str(err)
                    )
                    g.clog.warn(msg)
    returnValue(data)


def jsonFromFits(fname):
    hdr = fits.getheader(fname)

    def full_key(key):
        return "HIERARCH ESO {}".format(key)

    def get(name, default=None):
        return hdr.get(full_key(name), default)

    app_data = dict(
        multipliers=[1 + get("DET NSKIPS{}".format(i + 1), 0) for i in range(5)],
        dummy_out=get("DET DUMMY", 0),
        fastclk=get("DET FASTCLK", 0),
        oscan=int(get("DET INCPRSCX", False)),
        oscany=int(get("DET INCOVSCY", False)),
        readout="Slow" if get("DET SPEED", 0) == 0 else "Fast",
        xbin=get("DET BINX1", 1),
        ybin=get("DET BINY1", 1),
        clear=int(get("DET CLRCCD", True)),
        led_flsh=int(get("DET EXPLED", False)),
        dwell=get("DET TEXPOSE", 0.1) / 1000
        # TODO: numexp
    )

    user = dict(
        Observers=hdr.get("OBSERVER", ""),
        target=hdr.get("OBJECT", ""),
        comment=hdr.get("RUNCOM", ""),
        flags=hdr.get("IMAGETYP", "data"),
        filters=hdr.get("FILTERS", "us,gs,rs,is,zs"),
        ID=hdr.get("PROGRM", ""),
        PI=hdr.get("PI", ""),
    )

    mode = get("DET READ CURID")
    if mode == 2:
        # one window
        app_data["app"] = "Windows"
        app_data["x1size"] = get("DET WIN1 NX")
        app_data["y1size"] = get("DET WIN1 NY")
        app_data["x1start_lowerleft"] = get("DET WIN1 XSLL")
        app_data["x1start_lowerright"] = get("DET WIN1 XSLR")
        app_data["x1start_upperleft"] = get("DET WIN1 XSUL")
        app_data["x1start_upperright"] = get("DET WIN1 XSUR")
        app_data["y1start"] = get("DET WIN1 YS") + 1
    elif mode == 3:
        # two window
        app_data["app"] = "Windows"
        app_data["x1size"] = get("DET WIN1 NX")
        app_data["y1size"] = get("DET WIN1 NY")
        app_data["x1start_lowerleft"] = get("DET WIN1 XSLL")
        app_data["x1start_lowerright"] = get("DET WIN1 XSLR")
        app_data["x1start_upperleft"] = get("DET WIN1 XSUL")
        app_data["x1start_upperright"] = get("DET WIN1 XSUR")
        app_data["y1start"] = get("DET WIN1 YS") + 1
        app_data["x2size"] = get("DET WIN2 NX")
        app_data["y2size"] = get("DET WIN2 NY")
        app_data["x2start_lowerleft"] = get("DET WIN2 XSLL")
        app_data["x2start_lowerright"] = get("DET WIN2 XSLR")
        app_data["x2start_upperleft"] = get("DET WIN2 XSUL")
        app_data["x2start_upperright"] = get("DET WIN2 XSUR")
        app_data["y2start"] = get("DET WIN2 YS") + 1
    elif mode == 4:
        # drift mode
        app_data["app"] = "Drift"
        app_data["x1size"] = get("DET DRWIN NX")
        app_data["y1size"] = get("DET DRWIN NY")
        app_data["x1start_left"] = get("DET DRWIN XSL")
        app_data["x1start_right"] = get("DET DRWIN XSR")
        app_data["y1start"] = 1 + get("DET DRWIN YS")
    else:
        app_data["app"] = "FullFrame"

    setup_data = dict(appdata=app_data, user=user)
    return json.dumps(setup_data)


@inlineCallbacks
def insertFITSHDU(g):
    """
    Uploads a table of TCS data to the servers, which is appended onto a run.

    Arguments
    ---------
    g : hcam_drivers.globals.Container
        the Container object of application globals
    """
    if not g.cpars["hcam_server_on"]:
        g.clog.warn("insertFITSHDU: servers are not active")
        returnValue(False)

    session = g.session
    if session is None:
        g.clog.warn("no WAMP session")
        returnValue(False)

    run_number = yield getRunNumber(g)
    tcs_table = g.info.tcs_table

    g.clog.info("Adding TCS table data to run{:04d}.fits".format(run_number))
    try:
        fd = StringIO()
        ascii.write(tcs_table, format="ecsv", output=fd)
        yield session.call("hipercam.ngc.rpc.add_hdu", fd.getvalue(), run_number)
    except Exception as err:
        g.clog.warn("insertFITSHDU failed")
        msg = err.error_message() if hasattr(err, "error_message") else str(err)
        g.clog.warn(msg)

    nod_pattern = g.ipars.nodPattern
    if nod_pattern:
        g.clog.info("Adding nod pattern to run{:04d}.fits".format(run_number))
        try:
            nod_table = Table(nod_pattern)
            nod_table.rename_columns(["ra", "dec"], ["RAOFF", "DECOFF"])
            for colname in nod_table.colnames:
                nod_table[colname].unit = "arcsecond"
            fd = StringIO()
            ascii.write(nod_table, format="ecsv", output=fd)
            yield session.call("hipercam.ngc.rpc.add_hdu", fd.getvalue(), run_number)
        except Exception as err:
            g.clog.warn("insertFITSHDU failed")
            msg = err.error_message() if hasattr(err, "error_message") else str(err)
            g.clog.warn(msg)

    returnValue(True)


@inlineCallbacks
def execCommand(g, command, timeout=10):
    """
    Executes a command by sending it to the NGC server

    Arguments:
      g : hcam_drivers.globals.Container
        the Container object of application globals
      command : (string)
           the command (see below)

    Possible commands are:

      start              : starts a run
      stop               : stops a run
      abort              : aborts a run
      online             : bring ESO control server online and connect it to hardware
      offline            : put ESO control server in idle state and power down
      standby            : server can communicate, but child processes disabled
      reset              : resets the NGC controller front end

    Returns True/False according to whether the command
    succeeded or not.
    """
    if not g.cpars["hcam_server_on"]:
        g.clog.warn("execCommand: servers are not active")
        returnValue(False)

    session = g.session
    if session is None:
        g.clog.warn("no WAMP session")
        returnValue(False)

    try:
        response = yield session.call("hipercam.ngc.rpc.{}".format(command))
        if response is not None:
            msg, ok = response
        else:
            msg = ""
            ok = True
        g.clog.info('execCommand, command = "' + command + '"')

        if ok:
            g.clog.info("Response from server was OK")
            returnValue(True)
        else:
            g.clog.warn("Response from server was not OK")
            g.clog.warn("Reason: " + msg)
            returnValue(False)
    except Exception as err:
        g.clog.warn("execCommand failed")
        msg = err.error_message() if hasattr(err, "error_message") else str(err)
        g.clog.warn(msg)

    returnValue(False)


@inlineCallbacks
def isRunActive(g):
    """
    Polls the data server to see if a run is active
    """
    session = g.session if hasattr(g, "session") else None
    if session is None:
        g.clog.warn("no WAMP session")
        returnValue(False)

    if g.cpars["hcam_server_on"]:
        try:
            response = yield session.call("hipercam.ngc.rpc.summary")
        except Exception as err:
            msg = err.error_message() if hasattr(err, "error_message") else str(err)
            raise DriverError("isRunActive error reading NGC status: " + msg)
        tel = ReadNGCTelemetry(response)

        if not tel.ok:
            raise DriverError("isRunActive error: " + str(tel.err))
        if tel.state == "idle":
            returnValue(False)
        elif tel.state == "active":
            returnValue(True)
        elif tel.state == "error":
            msg = """
            NGC is in error state.
            Try resetting it using 'start_hicam' and 'Power On'
            """
            raise DriverError("isRunActive error\n" + msg)
        else:
            raise DriverError("isRunActive error, state = " + tel.state)
    else:
        raise DriverError("isRunActive error: servers are not active")


@inlineCallbacks
def isPoweredOn(g):
    session = g.session
    if session is None:
        g.clog.warn("no WAMP session")
        returnValue(False)

    if g.cpars["hcam_server_on"]:
        try:
            response = yield session.call("hipercam.ngc.rpc.summary")
        except Exception as err:
            msg = err.error_message() if hasattr(err, "error_message") else str(err)
            raise DriverError("isPoweredOn error reading NGC status: " + msg)

        tel = ReadNGCTelemetry(response)
        if not tel.ok:
            raise DriverError("isPoweredOn error: " + str(tel.err))
        if tel.clocks == "enabled":
            returnValue(True)
        else:
            returnValue(False)
    else:
        raise DriverError("isPoweredOn error: servers are not active")


@inlineCallbacks
def isOnline(g):
    # checks if ESO Server is in ONLINE state
    session = g.session
    if session is None:
        g.clog.warn("no WAMP session")
        returnValue(False)

    if g.cpars["hcam_server_on"]:
        try:
            msg, ok = yield session.call("hipercam.ngc.rpc.status")
        except Exception as err:
            msg = err.error_message() if hasattr(err, "error_message") else str(err)
            raise DriverError("isOnline error: " + msg)

        if not ok:
            raise DriverError("isOnline error: " + msg)
        if msg.lower() == "online":
            returnValue(True)
        else:
            returnValue(False)
    else:
        raise DriverError("isOnline error: hserver is not active")


@inlineCallbacks
def getFrameNumber(g):
    """
    Polls the data server to find the current frame number.

    Throws an exceotion if it cannot determine it.
    """
    session = g.session
    if session is None:
        g.clog.warn("no WAMP session")
        returnValue(False)

    if not g.cpars["hcam_server_on"]:
        raise DriverError("getRunNumber error: servers are not active")
    msg, ok = yield session.call("hipercam.ngc.rpc.status", "DET.FRAM2.NO")
    if not ok:
        raise DriverError("getFrameNumber error: could not get frame number " + msg)
    try:
        frame_no = int(msg)
    except ValueError:
        raise DriverError("getFrameNumber error: invalid msg " + msg)
    returnValue(frame_no)


@inlineCallbacks
def getRunNumber(g):
    """
    Polls the data server to find the current run number. Throws
    exceptions if it can't determine it.
    """
    session = g.session
    if session is None:
        g.clog.warn("no WAMP session")
        returnValue(False)

    if not g.cpars["hcam_server_on"]:
        raise DriverError("getRunNumber error: servers are not active")
    try:
        response = yield session.call("hipercam.ngc.rpc.summary")
    except Exception as err:
        msg = err.error_message() if hasattr(err, "error_message") else str(err)
        raise DriverError("isRunActive error reading NGC status: " + msg)
    tel = ReadNGCTelemetry(response)
    if tel.ok:
        returnValue(tel.run)
    else:
        raise DriverError("getRunNumber error: " + str(tel.err))


def checkSimbad(g, target, maxobj=5, timeout=5):
    """
    Sends off a request to Simbad to check whether a target is recognised.
    Returns with a list of results, or raises an exception if it times out
    """
    url = "http://simbad.u-strasbg.fr/simbad/sim-script"
    q = (
        "set limit "
        + str(maxobj)
        + '\nformat object form1 "Target: %IDLIST(1) | %COO(A D;ICRS)"\nquery '
        + target
    )
    query = urllib.parse.urlencode({"submit": "submit script", "script": q})
    resp = urllib.request.urlopen(url, query.encode(), timeout)
    data = False
    error = False
    results = []
    for line in resp:
        line = line.decode()
        if line.startswith("::data::"):
            data = True
        if line.startswith("::error::"):
            error = True
        if data and line.startswith("Target:"):
            name, coords = line[7:].split(" | ")
            results.append(
                {"Name": name.strip(), "Position": coords.strip(), "Frame": "ICRS"}
            )
    resp.close()

    if error and len(results):
        g.clog.warn(
            "drivers.check: Simbad: there appear to be some "
            + "results but an error was unexpectedly raised."
        )
    return results


class FifoThread(threading.Thread):
    """
    Adds a fifo Queue to a thread in order to store up disasters which are
    added to the fifo for later retrieval. This is to get around the problem
    that otherwise exceptions thrown from withins threaded operations are
    lost.
    """

    def __init__(self, name, target, fifo, args=()):
        threading.Thread.__init__(self, target=target, args=args)
        self.fifo = fifo
        self.name = name

    def run(self):
        """
        Version of run that traps Exceptions and stores
        them in the fifo
        """
        try:
            threading.Thread.run(self)
        except Exception:
            t, v, tb = sys.exc_info()
            error = traceback.format_exception_only(t, v)[0][:-1]
            tback = (
                self.name
                + " Traceback (most recent call last):\n"
                + "".join(traceback.format_tb(tb))
            )
            self.fifo.put((self.name, error, tback))



import six
import pickle
import json
import traceback
import itertools
from functools import partial
from astropy import units as u
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg

from twisted.internet.defer import inlineCallbacks

# internal imports
from ..misc import async_sleep
from ..mimic import Mimic
from ..tkutils import addStyle, get_root
from .utils import (
    plot_compo,
    INJECTOR_THETA,
    NOMINAL_PICKOFF_ZERO,
    NOMINAL_INJECTOR_ZERO,
    target_lens_position,
    PARK_POSITION,
    GUIDE_THETA,
    MAX_ANGLE,
)
from .. import widgets as w


if not six.PY3:
    import Tkinter as tk
else:
    import tkinter as tk


class COMPOSetupFrame(tk.Frame):
    """
    This is a minimal frame that contains only the buttons for injection side and the pickoff
    angle button.
    """

    def __init__(self, master):
        tk.Frame.__init__(self, master)
        addStyle(self)

        # create control widgets
        tk.Label(self, text="Injection Position").grid(
            row=0, column=0, pady=4, padx=4, sticky=tk.W
        )
        self.injection_side = w.Radio(
            self, ("L", "R", "G", "P"), 4, self.side_update, initial=1
        )
        self.injection_side.grid(row=0, column=1, pady=2, stick=tk.W)

        tk.Label(self, text="Pickoff Angle").grid(
            row=1, column=0, pady=4, padx=4, sticky=tk.W
        )
        self.pickoff_angle = w.RangedFloat(
            self,
            0.0,
            -MAX_ANGLE.to_value(u.deg),
            MAX_ANGLE.to_value(u.deg),
            None,
            False,
            allowzero=True,
            width=4,
        )
        self.pickoff_angle.grid(row=1, column=1, pady=2, stick=tk.W)

    @property
    def injection_angle(self):
        """
        A convenient property to convert positions to an angle

        Due to a nice coincidence, astropy units have a `value`
        attribute, as do w.RangedFloats. This means we can use
        `value` regardless of whether we are using a radio
        button to choose an injector angle, or a ranged float
        to set it manually.
        """
        if self.injection_side.value() == "L":
            ia = -INJECTOR_THETA
        elif self.injection_side.value() == "R":
            ia = INJECTOR_THETA
        elif self.injection_side.value() == "G":
            ia = GUIDE_THETA
        else:
            ia = PARK_POSITION
        return ia

    def side_update(self, *args):
        """
        Callback for injection side radio buttons

        Used to park pickoff arm if we are parking COMPO
        """
        if self.injection_side.value() == "P":
            self.pickoff_angle.set(-PARK_POSITION.to_value(u.deg))

    @property
    def lens_position(self):
        """
        A convenient property to find the lens position

        Due to a nice coincidence, astropy units have a `value`
        attribute, as do w.RangedFloats. This means we can use
        `value` regardless of whether we are using a radio
        button to choose an injector angle, or a ranged float
        to set it manually.
        """
        guiding = True if self.injection_side.value() == "G" else False
        return target_lens_position(
            abs(self.pickoff_angle.value()) * u.deg, guiding
        ).to(u.mm)


class COMPOSetupWidget(tk.Toplevel):
    """
    A child window to setup the COMPO pickoff arms.

    This is a minimal frame that contains only the buttons for injection side and the pickoff
    angle button. It is primarily used for defining instrument setups in hfinder.

    Normally this window is hidden, but can be revealed from the main GUIs menu
    or by clicking on a "use COMPO" widget in the main GUI.
    """

    def __init__(self, parent):
        tk.Toplevel.__init__(self, parent)
        self.transient(parent)
        self.parent = parent

        addStyle(self)
        self.title("COMPO setup")
        # do not display on creation
        self.withdraw()

        # dont destroy when we click the close button
        self.protocol("WM_DELETE_WINDOW", self.withdraw)

        self.setup_frame = COMPOSetupFrame(self)
        self.setup_frame.pack()

    def dumpJSON(self):
        """
        Encodes current COMPO setup data to JSON compatible dictionary
        """
        if self.setup_frame.injection_side.value() == "P":
            pickoff_angle = -PARK_POSITION.to_value(u.deg)
        else:
            pickoff_angle = self.setup_frame.pickoff_angle.value()
        return dict(
            injection_side=self.setup_frame.injection_side.value(),
            pickoff_angle=pickoff_angle,
        )

    def loadJSON(self, json_string):
        """
        Sets widget values from JSON data
        """
        data = json.loads(json_string)["compo"]
        self.setup_frame.injection_side.set(data["injection_side"])
        self.setup_frame.pickoff_angle.set(data["pickoff_angle"])


class CompoWidget(tk.Toplevel):
    """
    Parent class for COMPO widgets.

    Child classes need to implement the following attributes:
        - self.setup_frame: a COMPOSetupFrame, or duck-type equivalent
        - self.conn: a tk.Button that toggles connection to COMPO
        - self.injection_status: a w.Ilabel for status display
        - self.pickoff_status: a w.Ilabel for status display
        - self.lens_status: a w.Ilabel for status display
    """

    def __init__(self, parent):
        tk.Toplevel.__init__(self, parent)
        self.transient(parent)
        self.parent = parent

        addStyle(self)
        self.title("COMPO setup")

        # do not display on creation
        # self.withdraw()

        # dont destroy when we click the close button
        self.protocol("WM_DELETE_WINDOW", self.withdraw)

    @property
    def ok_to_start_run(self):
        """
        Convencience property to check if we are ready to start a run

        Runs should not start if any component is not in position
        """
        return (
            self.injection_status["text"] == "INPOS"
            and self.pickoff_status["text"] == "INPOS"
            and self.lens_status["text"] == "INPOS"
        )

    @property
    def session(self):
        return get_root(self).globals.session

    def get_value(self, thing, units):
        """
        Get a value either from a TK widget or a astropy unit

        This is a convenience function for getting values from either a
        tk widget or an astropy unit. We need it because attributes like
        lens_position or pickoff_angle can be either.

        Tk widgets have a "value" callable, whilst astropy units have a
        value attribute.

        Parameters
        ----------
        thing : object
            The object to get the value from
        units : astropy.units.Unit
            The units to convert to.

        Returns
        -------
        value : `astropy.units.Quantity`
            The value of the thing as an astropy quantity
        """
        try:
            value = thing.value() * units
        except TypeError:
            value = thing.to(units)
        return value

    @inlineCallbacks
    def handle_connection(self):
        if not self.session:
            self.print_message("no session")
            return

        rpc_templates = [
            "hipercam.compo_lens.rpc.connection.{}",
            "hipercam.compo_arms.rpc.connection.{}",
        ]
        # base action based on current value of the "conn" button
        # if label is "disconnect" then we want to disconnect...
        if self.conn["text"].lower() == "disconnect":
            rpcs = [template.format("disconnect") for template in rpc_templates]
            action = "disconnect"
        else:
            rpcs = [template.format("connect") for template in rpc_templates]
            action = "connect"
        try:
            for rpc in rpcs:
                yield self.session.call(rpc)
        except Exception as err:
            g = get_root(self).globals
            msg = err.error_message() if hasattr(err, "error_message") else str(err)
            g.clog.warn(f"Failed to {action} to COMPO: {msg}")

    @inlineCallbacks
    def home_stage(self, stage):
        if not self.session:
            self.print_message("no session")
            return
        if stage == "lens":
            rpc = "hipercam.compo_lens.rpc.stage.home"
        else:
            rpc = "hipercam.compo_arms.rpc.{}.home".format(stage)
        try:
            yield self.session.call(rpc)
        except Exception as err:
            g = get_root(self).globals
            msg = err.error_message() if hasattr(err, "error_message") else str(err)
            g.clog.warn(f"Failed to home {stage} in COMPO: {msg}")

    @inlineCallbacks
    def move_stage(self, stage):
        """
        Send commands to COMPO
        """
        if not self.session:
            self.print_message("no session")
            return

        if stage == "lens":
            position = self.get_value(self.setup_frame.lens_position, u.mm).value
            self.session.publish("hipercam.compo.target_lens_position", position)
        elif stage == "pickoff":
            position = self.get_value(self.setup_frame.pickoff_angle, u.deg).value
            self.session.publish("hipercam.compo.target_pickoff_angle", position)
        elif stage == "injection":
            position = self.get_value(self.setup_frame.injection_angle, u.deg).value
            self.session.publish("hipercam.compo.target_injection_angle", position)
        else:
            print(f"unrecognised stage: {stage}")
            return

        # allow time for statemachines to step forward (don't know why this is needed)
        yield async_sleep(1.5)
        if stage == "lens":
            rpc = "hipercam.compo_lens.rpc.stage.move"
        else:
            rpc = f"hipercam.compo_arms.rpc.{stage}.move"
        try:
            yield self.session.call(rpc)
        except Exception as err:
            g = get_root(self).globals
            msg = err.error_message() if hasattr(err, "error_message") else str(err)
            g.clog.warn(f"Failed to move {stage} in COMPO: {err}")

    @inlineCallbacks
    def stop_stage(self, stage):
        if not self.session:
            self.print_message("no session")
            return
        if stage == "lens":
            rpc = "hipercam.compo_lens.rpc.stage.stop"
        else:
            rpc = "hipercam.compo_arms.rpc.{}.stop".format(stage)
        try:
            yield self.session.call(rpc)
        except Exception as err:
            g = get_root(self).globals
            msg = err.error_message() if hasattr(err, "error_message") else str(err)
            g.clog.warn(f"Failed to stop {stage} in COMPO: {msg}")

    def update_target_positions(self):
        """
        This does not move the stages, but simply updates the target positions
        """
        if not self.session:
            self.print_message("no session")
            return

        ia = self.get_value(self.setup_frame.injection_angle, u.deg)
        ia += NOMINAL_INJECTOR_ZERO

        poa = self.get_value(self.setup_frame.pickoff_angle, u.deg)
        poa += NOMINAL_PICKOFF_ZERO

        lens = self.get_value(self.setup_frame.lens_position, u.mm).value

        self.session.publish("hipercam.compo.target_pickoff_angle", poa.to_value(u.deg))
        self.session.publish(
            "hipercam.compo.target_injection_angle", ia.to_value(u.deg)
        )
        self.session.publish("hipercam.compo.target_lens_position", lens)

    @inlineCallbacks
    def move(self):
        """
        Send commands to COMPO.

        This is more efficient than calling CompoWidget.move_stage for each stage
        """
        if not self.session:
            self.print_message("no session")
            return

        self.update_target_positions()
        # allow time for statemachines to step forward
        yield async_sleep(1.5)

        try:
            yield self.session.call("hipercam.compo_arms.rpc.pickoff.move")
            yield self.session.call("hipercam.compo_arms.rpc.injection.move")
            yield self.session.call("hipercam.compo_lens.rpc.stage.move")
        except Exception as err:
            g = get_root(self).globals
            msg = err.error_message() if hasattr(err, "error_message") else str(err)
            g.clog.warn(f"Failed to move stages in COMPO: {msg}")

    @inlineCallbacks
    def home_all(self):
        for stage in ("injection", "pickoff", "lens"):
            yield self.home_stage(stage)

    @inlineCallbacks
    def stop_all(self):
        for stage in ("injection", "pickoff", "lens"):
            yield self.stop_stage(stage)

    def send_message(self, topic, msg):
        if self.session:
            self.session.publish(topic, msg)

    def print_message(self, msg):
        # put message inside label widget
        self.label.delete(1.0, tk.END)
        self.label.insert(tk.END, msg + "\n")

    def set_stage_status(self, stage, telemetry):
        if stage == "lens":
            state = telemetry["state"]["lens_state"]["stage"]
        else:
            state = telemetry["state"]["arms_state"][stage]

        if stage == "injection":
            widget = self.injection_status
        elif stage == "pickoff":
            widget = self.pickoff_status
        elif stage == "lens":
            widget = self.lens_status
        else:
            raise ValueError("unkown stage " + stage)

        g = get_root(self).globals
        colours = {
            "inpos": g.COL["start"],
            "outpos": g.COL["warn"],
            "moving": g.COL["warn"],
            "stopped": g.COL["warn"],
            "init": g.COL["warn"],
            "homing": g.COL["warn"],
            "disabled": g.COL["warn"],
        }
        matched_state = set(state).intersection(colours.keys())
        if not matched_state:
            print("unhandled state " + "/".join(state))
        elif len(matched_state) > 1:
            print("ambiguous state " + "/".join(state))
        else:
            state = matched_state.pop()
            c = colours[state]
            widget.config(text=state.upper(), bg=c)

    def update_mimic(self, telemetry):
        """
        Use incoming telemetry to update mimic
        """
        if not telemetry:
            return

        injection_angle, _ = (
            self.get_stage_position(telemetry, "injection_angle") * u.deg
        )
        pickoff_angle, _ = self.get_stage_position(telemetry, "pickoff_angle") * u.deg

        injection_angle -= NOMINAL_INJECTOR_ZERO
        pickoff_angle -= NOMINAL_PICKOFF_ZERO

        self.ax.clear()
        _ = plot_compo(pickoff_angle, injection_angle, self.ax)
        self.ax.set_xlim(-250, 250)
        self.ax.set_aspect("equal")
        self.ax.set_axis_off()
        self.canvas.draw()

    def get_stage_position(self, telemetry, pos_str):
        try:
            pos = telemetry[pos_str]["current"].value
        except AttributeError:
            pos = telemetry[pos_str]["current"]

        try:
            targ = telemetry[pos_str]["target"].value
        except AttributeError:
            targ = telemetry[pos_str]["target"]
        return pos, targ

    def on_telemetry(self, package_data):
        try:
            telemetry = pickle.loads(package_data)
            state = telemetry["state"]

            # check for error status
            g = get_root(self).globals
            # extract connection states from telemetry state package
            # and join into one long list for both arms and lens
            connection_states = list(
                itertools.chain(*[s["connection"] for s in state.values()])
            )
            if "error" in connection_states:
                self.lens_status.config(text="ERROR", bg=g.COL["critical"])
                self.pickoff_status.config(text="ERROR", bg=g.COL["critical"])
                self.injection_status.config(text="ERROR", bg=g.COL["critical"])
                self.conn.config(text="Connect")
            elif "offline" in connection_states:
                self.lens_status.config(text="DISCONN", bg=g.COL["critical"])
                self.pickoff_status.config(text="DISCONN", bg=g.COL["critical"])
                self.injection_status.config(text="DISCONN", bg=g.COL["critical"])
                self.conn.config(text="Connect")
            else:
                self.conn.config(text="Disconnect")
                for stage in ("injection", "pickoff", "lens"):
                    self.set_stage_status(stage, telemetry)

            str = f"{telemetry['timestamp'].iso}:\n"
            for key, stage, pos_str in zip(
                ("arms_state", "arms_state", "lens_state"),
                ("injection", "pickoff", "stage"),
                ("injection_angle", "pickoff_angle", "lens_position"),
            ):
                pos, targ = self.get_stage_position(telemetry, pos_str)

                status = "/".join(state[key][stage][4:])
                str += f"{stage}: curr={pos:.2f}, targ={targ:.2f}\n{status}\n\n"

            self.print_message(str)
            try:
                self.update_mimic(telemetry)
            except ValueError:
                # can sometimes fail if we are missing one or more angles
                pass
        except Exception as err:
            print("error handling COMPO telemetry")
            print(traceback.format_exc())

    def dumpJSON(self):
        """
        Encodes current COMPO setup data to JSON compatible dictionary
        """
        if self.setup_frame.injection_side.value() == "P":
            pickoff_angle = -PARK_POSITION.to_value(u.deg)
        else:
            pickoff_angle = self.setup_frame.pickoff_angle.value()
        return dict(
            injection_side=self.setup_frame.injection_side.value(),
            pickoff_angle=pickoff_angle,
        )

    def loadJSON(self, json_string):
        """
        Sets widget values from JSON data
        """
        try:
            data = json.loads(json_string)["compo"]
        except KeyError:
            # no compo data in JSON, park arms
            data = dict(
                injection_side="P",
                pickoff_angle=-PARK_POSITION.to_value(u.deg),
            )
        self.setup_frame.injection_side.set(data["injection_side"])
        self.setup_frame.pickoff_angle.set(data["pickoff_angle"])
        self.update_target_positions()


class COMPOControlWidget(CompoWidget):
    """
    A child window to control the COMPO pickoff arms.

    This is a more advanced window that adds widgets to monitor the state of COMPO
    and allow user control of the hardware. It is used in hdriver.

    Normally this window is hidden, but can be revealed from the main GUIs menu
    or by clicking on a "use COMPO" widget in the main GUI.
    """

    def __init__(self, parent):
        CompoWidget.__init__(self, parent)

        g = get_root(self).globals
        # frames for sections
        left = tk.Frame(self)
        right = tk.Frame(self)

        self.setup_frame = COMPOSetupFrame(left)
        self.setup_frame.grid(row=1, column=0, columnspan=2, pady=2, sticky=tk.W)

        # buttons
        self.go = w.ActButton(left, width=12, callback=self.move, text="Move")
        self.conn = w.ActButton(
            left, width=12, callback=self.handle_connection, text="Connect"
        )
        self.stop = w.ActButton(left, width=12, callback=self.stop_all, text="Stop")
        self.home = w.ActButton(left, width=12, callback=self.home_all, text="Home")

        self.conn.grid(row=2, column=0, pady=2, sticky=tk.E)
        self.home.grid(row=2, column=1, pady=2, sticky=tk.W)
        self.stop.grid(row=3, column=1, pady=2, sticky=tk.W)
        self.go.grid(row=3, column=0, pady=2, sticky=tk.E)

        # create status widgets
        status = tk.LabelFrame(left, text="status")
        status.grid(row=4, column=0, columnspan=2, pady=4, padx=4, sticky=tk.N)

        tk.Label(status, text="Injection Arm").grid(row=0, column=0, sticky=tk.W)
        self.injection_status = w.Ilabel(status, text="INIT", width=10, anchor=tk.W)
        self.injection_status.config(bg=g.COL["warn"])
        self.injection_status.grid(row=0, column=1, sticky=tk.W, pady=2, padx=2)

        tk.Label(status, text="Pickoff Arm").grid(row=1, column=0, sticky=tk.W)
        self.pickoff_status = w.Ilabel(status, text="INIT", width=10, anchor=tk.W)
        self.pickoff_status.config(bg=g.COL["warn"])
        self.pickoff_status.grid(row=1, column=1, sticky=tk.W, pady=2, padx=2)

        tk.Label(status, text="Lens Position").grid(row=2, column=0, sticky=tk.W)
        self.lens_status = w.Ilabel(status, text="INIT", width=10, anchor=tk.W)
        self.lens_status.config(bg=g.COL["warn"])
        self.lens_status.grid(row=2, column=1, sticky=tk.W, pady=2, padx=2)

        # telemetry
        tel_frame = tk.LabelFrame(right, text="telemetry")
        self.label = tk.Text(tel_frame, height=10, width=40, bg=g.COL["log"])
        self.label.configure(state=tk.NORMAL, font=g.ENTRY_FONT)
        self.label.pack(fill=tk.Y)
        tel_frame.grid(row=0, column=0, columnspan=1)

        # mimic
        mimic_width = 350
        Mimic.__init__(self, height=int(mimic_width / 2.5), width=mimic_width)
        mimic_frame = tk.LabelFrame(right, text="mimic")
        self.canvas = FigureCanvasTkAgg(self.figure, mimic_frame)
        self.canvas.get_tk_widget().pack()
        mimic_frame.grid(row=1, column=0, padx=4, pady=4)

        left.pack(pady=2, side=tk.LEFT, fill=tk.Y)
        right.pack(pady=2, side=tk.LEFT, fill=tk.Y)


class COMPOManualWidget(CompoWidget):
    """
    This is a child window to manually control COMPO.

    This window just allows you to manually set the positions of the arms and slide.
    It also allows independent homing/stopping of each device. It is used by the
    bespoke COMPO GUI script in hcam_drivers, which is designed for full manual control.

    It has `injection_angle`, `pickoff_angle`, and `lens_position` widgets that
    are w.RangedFloats. This allows the widget to be used as duck-typed drop-in for
    a COMPOSetupFrame.
    """

    def __init__(self, parent):
        CompoWidget.__init__(self, parent)
        g = get_root(self).globals

        # connection button
        self.conn = w.ActButton(
            self, width=12, callback=self.handle_connection, text="Connect"
        )
        self.conn.grid(row=0, column=0, columnspan=2, pady=2, sticky=tk.E)

        # pickoff
        row = 1
        tk.Label(self, text="Pickoff Angle (deg)").grid(
            row=row, column=0, pady=4, padx=4, sticky=tk.W
        )
        self.pickoff_angle = w.RangedFloat(
            self, 0.0, -67, 67, None, False, allowzero=True, width=4
        )
        self.pickoff_angle.grid(row=row, column=1, pady=2, stick=tk.W)
        self.pickoff_home = w.ActButton(
            self, width=12, callback=partial(self.home_stage, "pickoff"), text="Home"
        )
        self.pickoff_home.grid(row=row, column=2, pady=2, stick=tk.W)
        self.pickoff_move = w.ActButton(
            self, width=12, callback=partial(self.move_stage, "pickoff"), text="Move"
        )
        self.pickoff_move.grid(row=row, column=3, pady=2, stick=tk.W)
        self.pickoff_stop = w.ActButton(
            self, width=12, callback=partial(self.stop_stage, "pickoff"), text="Stop"
        )
        self.pickoff_stop.grid(row=row, column=4, pady=2, stick=tk.W)

        # injection arm
        row = 2
        tk.Label(self, text="Injection Angle (deg)").grid(
            row=row, column=0, pady=4, padx=4, sticky=tk.W
        )
        self.injection_angle = w.RangedFloat(
            self, 0.0, -67, 67, None, False, allowzero=True, width=4
        )
        self.injection_angle.grid(row=row, column=1, pady=2, stick=tk.W)
        self.injection_home = w.ActButton(
            self, width=12, callback=partial(self.home_stage, "injection"), text="Home"
        )
        self.injection_home.grid(row=row, column=2, pady=2, stick=tk.W)
        self.injection_move = w.ActButton(
            self, width=12, callback=partial(self.move_stage, "injection"), text="Move"
        )
        self.injection_move.grid(row=row, column=3, pady=2, stick=tk.W)
        self.injection_stop = w.ActButton(
            self, width=12, callback=partial(self.stop_stage, "injection"), text="Stop"
        )
        self.injection_stop.grid(row=row, column=4, pady=2, stick=tk.W)

        # lens
        row = 3
        tk.Label(self, text="Lens (mm)").grid(
            row=row, column=0, pady=4, padx=4, sticky=tk.W
        )
        self.lens_position = w.RangedFloat(
            self, 0.0, 0, 25, None, False, allowzero=True, width=4
        )
        self.lens_position.grid(row=row, column=1, pady=2, stick=tk.W)
        self.lens_home = w.ActButton(
            self, width=12, callback=partial(self.home_stage, "lens"), text="Home"
        )
        self.lens_home.grid(row=row, column=2, pady=2, stick=tk.W)
        self.lens_move = w.ActButton(
            self, width=12, callback=partial(self.move_stage, "lens"), text="Move"
        )
        self.lens_move.grid(row=row, column=3, pady=2, stick=tk.W)
        self.lens_stop = w.ActButton(
            self, width=12, callback=partial(self.stop_stage, "lens"), text="Stop"
        )
        self.lens_stop.grid(row=row, column=4, pady=2, stick=tk.W)

        # create status widgets
        status = tk.LabelFrame(self, text="status")
        status.grid(row=4, column=0, columnspan=4, pady=4, padx=4, sticky=tk.N)

        tk.Label(status, text="Injection Arm").grid(row=0, column=0, sticky=tk.W)
        self.injection_status = w.Ilabel(status, text="INIT", width=10, anchor=tk.W)
        self.injection_status.config(bg=g.COL["warn"])
        self.injection_status.grid(row=0, column=1, sticky=tk.W, pady=2, padx=2)

        tk.Label(status, text="Pickoff Arm").grid(row=row, column=0, sticky=tk.W)
        self.pickoff_status = w.Ilabel(status, text="INIT", width=10, anchor=tk.W)
        self.pickoff_status.config(bg=g.COL["warn"])
        self.pickoff_status.grid(row=row, column=1, sticky=tk.W, pady=2, padx=2)

        tk.Label(status, text="Lens Position").grid(row=2, column=0, sticky=tk.W)
        self.lens_status = w.Ilabel(status, text="INIT", width=10, anchor=tk.W)
        self.lens_status.config(bg=g.COL["warn"])
        self.lens_status.grid(row=2, column=1, sticky=tk.W, pady=2, padx=2)

        # telemetry
        tel_frame = tk.LabelFrame(self, text="telemetry")
        self.label = tk.Text(tel_frame, height=10, width=40, bg=g.COL["log"])
        self.label.configure(state=tk.NORMAL, font=g.ENTRY_FONT)
        self.label.pack(fill=tk.Y)
        tel_frame.grid(row=5, column=0, columnspan=4, pady=4, padx=4, sticky=tk.N)

        # mimic (not shown)
        mimic_width = 350
        Mimic.__init__(self, height=int(mimic_width / 2.5), width=mimic_width)
        mimic_frame = tk.LabelFrame(self, text="mimic")
        self.canvas = FigureCanvasTkAgg(self.figure, mimic_frame)

    @property
    def setup_frame(self):
        return self



from __future__ import print_function, division, unicode_literals, absolute_import
from six.moves.queue import Queue

"""
This module supplies a single class, the Container. The idea is
that the root widget (the GUI mainwindow) has a single Container
property. Since all sub-widgets can get a reference to the root
window, all sub widgets can access the Container and thus communicate,
hence avoiding the use of globals.
"""


class Container:
    """
    A simple class to hold common attributes for communication between widgets.

    Those in CAPITALS are meant to be immutable; lowercase ones get
    updated.

    The meaning of the globals is as follows (uppercase)::

    COL   : colours
    DAY   : number of seconds in a day
    EFAC  : ratio FWHM/sigma for a gaussian
    EXTINCTION : extinction, mags/airmass
    SKY   : sky brightenesses
    TINS  : Telescop/instrument data
    DEFAULT_FONT : standard font (set in drivers)
    MENU_FONT    : font for menus (set in drivers)
    ENTRY_FONT   : font for data entry fields and mutable fields

    SERVER_POST_PATH : url for posting setups to rack server

    and lowercase (all set = None to start with)::

    astro   : astronomical information
    clog    : command log widget. Used to report actions and results
    count   : count & S/N information widget
    cpars   : dictionary of configuration parameters
    fpslide : focal plane slide widget
    info    : information widget
    ipars   : instrument parameters widget (windows sizes etc)
    observe : widget of observing commands
    rlog    : response log widget. Used to report server responses
    rpars   : run parameter widget
    setup   : setup widget
    ccd_hw  : CCD hardware (temps, flow rates etc) monitoring widget
    start_filter : filter at last press of 'Start'
    logfile : file to log messages
    """

    def __init__(self, telescope_names=None):
        # Ratio of FWHM/sigma for a gaussian
        self.EFAC = 2.3548

        # FIFO Queue for retrieving exceptions from
        # threaded operations
        self.FIFO = Queue()

        # Colours
        self.COL = {
            "main": "#d0d0ff",  # Colour for the surrounds
            "text": "#000050",  # Text colour
            "debug": "#a0a0ff",  # Text background for debug messages
            "warn": "#f0c050",  # Text background for warnings
            "error": "#ffa0a0",  # Text background for errors
            "critical": "#ff0000",  # Text background for disasters
            "start": "#00e000",  # Start / Success button colour when enabled
            "stop": "#ff5050",  # Stop / Failure button colour when enabled
            "startD": "#d0e0d0",  # Start / Success button colour when disabled
            "stopD": "#ffe0e0",  # Stop / Failure button colour when disabled
            "log": "#e0d4ff",  # Logger windows
        }

        # Telescope / instrument info. Most of this is do with estimating count rates
        self.TINS = {
            "WHT": {
                "latitude": 28.7598742,  # latitude degrees, North positive
                "longitude": -17.8793802,  # longitude degrees, East positive
                "elevation": 2327.0,  # Elevation above sea level, metres
                "plateScale": 0.3,  # Arcsecs/unbinned pixel
                "zerop": {"u": 25.67, "g": 27.25, "r": 26.65, "i": 26.79, "z": 26.39},
            },
            "GTC": {
                "latitude": 28.762,  # latitude degrees, North positive
                "longitude": -17.87764,  # longitude degrees, East positive
                "elevation": 2300.0,  # Elevation above sea level, metres
                "plateScale": 0.081,  # Arcsecs/unbinned pixel (WHT Collimator)
                "zerop": {
                    "u": 27.47,
                    "g": 29.06,
                    "r": 28.45,
                    "i": 28.20,
                    "z": 27.86,
                },
            },
            "TNT": {
                "latitude": 18.574,  # latitude degrees, North positive
                "longitude": 98.482,  # longitude degrees, East positive
                "elevation": 2449.0,  # Elevation above sea level, metres
                "plateScale": 0.456,  # Arcsecs/unbinned pixel (WHT Collimator)
                "zerop": {"u": 22.71, "g": 25.25, "r": 25.01, "i": 24.69, "z": 23.81},
            },
            "NTT_CUBE": {
                "latitude": -29.256,  # latitude degrees, North positive
                "longitude": -70.73,  # longitude degrees, East positive
                "elevation": 2347.0,  # Elevation above sea level, metres
                "plateScale": 0.354,  # Arcsecs/unbinned pixel (WHT Collimator)
                "zerop": {"u": 24.62, "g": 26.43, "r": 25.77, "i": 25.63, "z": 24.79},
            },
        }
        if telescope_names is not None:
            self.TINS = {k: self.TINS[k] for k in telescope_names}

        # Sky brightness, mags/sq-arsec
        self.SKY = {
            "d": {"u": 22.4, "g": 22.2, "r": 21.4, "i": 20.7, "z": 20.3},
            "g": {"u": 21.4, "g": 21.2, "r": 20.4, "i": 20.1, "z": 19.9},
            "b": {"u": 18.4, "g": 18.2, "r": 17.4, "i": 17.9, "z": 18.3},
        }

        # Extinction per unit airmass
        self.EXTINCTION = {"u": 0.5, "g": 0.19, "r": 0.09, "i": 0.05, "z": 0.04}

        # Fonts set by drivers.add_style later
        # Default font, e.g. used for fixed labels
        self.DEFAULT_FONT = None

        # Font for menus
        self.MENU_FONT = None

        # Entry font, used for data entry points and mutable
        # information
        self.ENTRY_FONT = None

        # URL for posting setups
        self.SERVER_POST_PATH = "setup"

        # Command log widget. Used to report actions and results
        self.clog = None

        # Response log widget. Used to report server responses
        self.rlog = None

        # Configuration parameter dictionary: configurable options
        self.cpars = None

        # Instrument parameters widget
        self.ipars = None

        # Run parameters widget
        self.rpars = None

        # Astronomical information widget
        self.astro = None

        # Information widget
        self.info = None

        # Focal plane slide widget
        self.fpslide = None

        # Observing widget
        self.observe = None

        # Count rates, S-to-N etc widget
        self.count = None

        # Instrument setup widget
        self.setup = None

        # ccd hw monitor widget
        self.ccd_hw = None

        # Logging file
        self.logfile = None



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# This sub-module is destined for common non-package specific utility
# functions that will ultimately be merged into `astropy.utils`

# As far as possible, utils contains classes of generic use, such as
# PosInt for positive integer input. See e.g. 'hcam' for more instrument
# dependent components.

from __future__ import print_function, unicode_literals, absolute_import, division
import six

if not six.PY3:
    import tkFont as font
else:
    from tkinter import font


def get_root(w):
    """
    Simple method to access root for a widget
    """
    next_level = w
    while next_level.master:
        next_level = next_level.master
    return next_level


def addStyle(w):
    """
    Styles the GUI: global fonts and colours.

    Parameters
    ----------
    w : tkinter.tk
        widget element to style
    """
    # access global container in root widget
    root = get_root(w)
    g = root.globals
    fsize = g.cpars["font_size"]
    family = g.cpars["font_family"]

    # Default font
    g.DEFAULT_FONT = font.nametofont("TkDefaultFont")
    g.DEFAULT_FONT.configure(size=fsize, weight="bold", family=family)
    w.option_add("*Font", g.DEFAULT_FONT)

    # Menu font
    g.MENU_FONT = font.nametofont("TkMenuFont")
    g.MENU_FONT.configure(family=family)
    w.option_add("*Menu.Font", g.MENU_FONT)

    # Entry font
    g.ENTRY_FONT = font.nametofont("TkTextFont")
    g.ENTRY_FONT.configure(size=fsize, family=family)
    w.option_add("*Entry.Font", g.ENTRY_FONT)

    # position and size
    # root.geometry("320x240+325+200")

    # Default colours. Note there is a difference between
    # specifying 'background' with a capital B or lowercase b
    w.option_add("*background", g.COL["main"])
    w.option_add("*HighlightBackground", g.COL["main"])
    w.config(background=g.COL["main"])


def place_at_edge(parent, win, padding=10):
    """
    Places a window at the edge of its parent window.

    This is useful when deiconifying a window that has been withdrawn
    so that it doesn't overlap with the main window.

    Parameters
    ----------
    parent : tkinter widget
        parent widget
    win : tkinter widget
        widget to place
    padding : int
        padding in pixels between windows
    """

    # make sure windows are updated so we get current positions
    win.update_idletasks()
    parent.update_idletasks()

    # width and height of window to place
    width = win.winfo_width()
    height = win.winfo_height()

    # position and size of parent
    parent_x_upperleft = parent.winfo_rootx()
    parent_y_upperleft = parent.winfo_rooty()
    titlebar_height = parent.winfo_rooty() - parent.winfo_y()
    parent_width = parent.winfo_width()

    win.geometry(
        "{}x{}+{}+{}".format(
            width,
            height,
            parent_x_upperleft + parent_width + padding,
            parent_y_upperleft - titlebar_height,
        )
    )
